Package: CKNNRLD
Title: Clustering-Based K-Nearest Neighbor Regression for Longitudinal
        Data
Version: 0.1.4
Authors@R: c(
    person(given = "Mohammad Sadegh", family = "Loeloe", email = "mslbiostat@gmail.com", role = c("aut", "cre")),
    person(given = "Seyyed Mohammad", family = "Tabatabaei", role = "aut"),
    person(given = "Reyhane", family = "Sefidkar", role = "aut"),
    person(given = "Amir Houshang", family = "Mehrparvar", role = "aut"),
    person(given = "Sara", family = "Jambarsang", role = c("aut", "ths"))
  )
Description: Implements the 'CKNNRLD' algorithm (Clustering-Based K-Nearest
    Neighbor Regression for Longitudinal Data) for improving K-Nearest
    Neighbor ('KNN') regression on longitudinal data through cluster-based
    partitioning and localized prediction. Offers enhanced computational
    efficiency and accuracy for high-volume longitudinal datasets. The
    acronym 'KNN' stands for K-Nearest Neighbor. References: Loeloe MS,
    Tabatabaei SM, Sefidkar R, Mehrparvar AH, Jambarsang S (2025).
    "Boosting K-nearest neighbor regression performance for longitudinal
    data through a novel learning approach." BMC Bioinformatics, 26, 232.
    <doi:10.1186/s12859-025-06205-1>.
License: GPL-3
Encoding: UTF-8
Imports: Directional, graphics, Rfast
Depends: R (>= 3.5.0)
NeedsCompilation: no
Language: en-US
Config/roxygen2/version: 8.0.0
Packaged: 2026-06-09 16:08:47 UTC; sadegh-pc
Author: Mohammad Sadegh Loeloe [aut, cre],
  Seyyed Mohammad Tabatabaei [aut],
  Reyhane Sefidkar [aut],
  Amir Houshang Mehrparvar [aut],
  Sara Jambarsang [aut, ths]
Maintainer: Mohammad Sadegh Loeloe <mslbiostat@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-09 16:20:02 UTC
Built: R 4.4.3; ; 2026-06-09 16:56:25 UTC; unix

# Generated by roxygen2: do not edit by hand

export(BestC)
export(CKNNRLD)
export(CKNNRLD.tune)
export(KNNRLD)
export(KNNRLD.tune)
importFrom(Directional,makefolds)
importFrom(Rfast,colmeans)
importFrom(Rfast,dista)
importFrom(graphics,abline)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(stats,kmeans)

#' Tune CKNNRLD Model with Automatic Cluster Selection
#'
#' Automatically selects the best number of clusters (C) using internal validation
#' criteria, then tunes the CKNNRLD model within each cluster via cross-validation.
#'
#' @param y Matrix of longitudinal outcomes (subjects x timepoints).
#' @param x Matrix of predictor variables (subjects x features).
#' @param nfolds Number of folds for cross-validation (default = 10).
#' @param folds Optional list of pre-specified fold indices.
#' @param seed Random seed for reproducibility.
#' @param A Maximum number of neighbors to evaluate (searches from 2 to A, default = 10).
#' @param C_range Range of cluster numbers to evaluate (default = 2:6).
#' @param cluster_method Clustering method to use. Currently supports "kml" (default).
#'
#' @return A list containing:
#' \item{best_c}{Optimal number of clusters}
#' \item{cluster_results}{Tuning results for each cluster (from KNNRLD.tune)}
#' \item{cluster_sizes}{Size of each cluster}
#' \item{cluster_assignments}{Cluster labels for each subject}
#' \item{criteria}{Data frame of quality criteria used for selecting best C}
#'
#' @examples
#' \donttest{
#' set.seed(123)
#' n <- 30
#' T <- 3
#' d <- 2
#' x <- matrix(runif(n * d), nrow = n)
#' y <- matrix(rnorm(n * T), nrow = n)
#' tune_result <- CKNNRLD.tune(
#'   y = y,
#'   x = x,
#'   nfolds = 3,
#'   A = 4,
#'   C_range = 2:3
#' )
#' print(tune_result$best_c)
#' }
#'
#' @importFrom latrend lcMethodKML latrend trajectoryAssignments
#' @importFrom stats reshape logLik AIC BIC
#'
#' @export
CKNNRLD.tune <- function(y, x, nfolds = 10, folds = NULL, seed = NULL, 
                          A = 10, C_range = 2:6, cluster_method = "kml") {
  
  y <- as.matrix(y)
  x <- as.matrix(x)
  
  n_subjects <- nrow(y)
  n_timepoints <- ncol(y)
  n_predictors <- ncol(x)
  
  long_data <- data.frame(
    id = 1:n_subjects,
    time = rep(1:n_timepoints, each = n_subjects),
    value = as.vector(t(y))
  )
  
  calc_calinski <- function(data, clusters) {
    k <- length(unique(clusters))
    n <- nrow(data)
    if (k < 2 || n <= k) return(NA)
    overall_mean <- colMeans(data)
    between_ss <- 0
    for (i in 1:k) {
      cluster_data <- data[clusters == i, , drop = FALSE]
      if (nrow(cluster_data) > 0) {
        cluster_mean <- colMeans(cluster_data)
        between_ss <- between_ss + nrow(cluster_data) * sum((cluster_mean - overall_mean)^2)
      }
    }
    within_ss <- 0
    for (i in 1:k) {
      cluster_data <- data[clusters == i, , drop = FALSE]
      if (nrow(cluster_data) > 0) {
        cluster_mean <- colMeans(cluster_data)
        within_ss <- within_ss + sum(rowSums((cluster_data - cluster_mean)^2))
      }
    }
    if (within_ss == 0) return(NA)
    return((between_ss / (k - 1)) / (within_ss / (n - k)))
  }
  
  crit_list <- list()
  
  for (k in C_range) {
    if (cluster_method == "kml") {
      method_obj <- lcMethodKML(
        response = "value",
        time = "time",
        id = "id",
        nClusters = k,
        save.data = FALSE
      )
    } else {
      stop(paste("Method", cluster_method, "is not implemented. Use 'kml'."))
    }
    
    model <- tryCatch({
      latrend(method_obj, data = long_data, verbose = FALSE)
    }, error = function(e) {
      warning(paste("Clustering failed for C =", k, ":", e$message))
      return(NULL)
    })
    
    if (!is.null(model)) {
      cluster_assign <- trajectoryAssignments(model)
      traj_data <- as.data.frame(model)
      traj_wide <- stats::reshape(traj_data, 
                                  idvar = "id", 
                                  timevar = "time", 
                                  direction = "wide")
      traj_matrix <- as.matrix(traj_wide[, grep("value", names(traj_wide))])
      colnames(traj_matrix) <- NULL
      calinski_val <- calc_calinski(traj_matrix, cluster_assign)
      criteria <- c(
        CalinskiHarabasz = calinski_val,
        LogLik = as.numeric(stats::logLik(model)),
        AIC = stats::AIC(model),
        BIC = stats::BIC(model)
      )
      crit_list[[paste0("c", k)]] <- criteria
    } else {
      crit_list[[paste0("c", k)]] <- rep(NA, 4)
    }
  }
  
  max_len <- max(sapply(crit_list, length))
  crit_list_padded <- lapply(crit_list, function(x) {
    if (length(x) < max_len) c(x, rep(NA, max_len - length(x))) else x
  })
  
  criteria_df <- as.data.frame(do.call(cbind, crit_list_padded))
  rownames(criteria_df) <- c("CalinskiHarabasz", "LogLik", "AIC", "BIC")
  
  calinski_row <- which(rownames(criteria_df) == "CalinskiHarabasz")
  if (length(calinski_row) > 0 && any(!is.na(criteria_df[calinski_row, ]))) {
    best_c <- C_range[which.max(criteria_df[calinski_row, ])]
  } else {
    best_c <- C_range[1]
    warning("Could not determine optimal clusters. Using first C value.")
  }
  
  final_method <- lcMethodKML(
    response = "value",
    time = "time",
    id = "id",
    nClusters = best_c,
    save.data = FALSE
  )
  
  final_model <- latrend(final_method, data = long_data, verbose = FALSE)
  cluster_assignments <- trajectoryAssignments(final_model)
  
  tdata <- cbind(cluster = cluster_assignments, y, x)
  cluster_sizes <- table(cluster_assignments)
  
  y_start_idx <- 2
  y_end_idx <- y_start_idx + n_timepoints - 1
  x_start_idx <- y_end_idx + 1
  x_end_idx <- x_start_idx + n_predictors - 1
  
  xnew_clustered <- cbind(cluster = cluster_assignments, x)
  
  cluster_results <- vector("list", best_c)
  
  for (i in 1:best_c) {
    cluster_rows <- which(xnew_clustered[, 1] == i)
    
    if (length(cluster_rows) == 0) {
      warning(paste("No data for Cluster =", i))
      cluster_results[[i]] <- NULL
      next
    }
    
    x_i <- tdata[tdata[, 1] == i, x_start_idx:x_end_idx, drop = FALSE]
    y_i <- tdata[tdata[, 1] == i, y_start_idx:y_end_idx, drop = FALSE]
    
    if (nrow(y_i) < 3) {
      warning(paste("Cluster", i, "has insufficient data (n =", nrow(y_i), "). Skipping tuning."))
      cluster_results[[i]] <- list(
        crit = NA,
        best_k = NA,
        performance = NA,
        warning = "Insufficient data"
      )
      next
    }
    
    cluster_results[[i]] <- KNNRLD.tune(
      y = y_i,
      x = x_i,
      nfolds = nfolds,
      folds = folds,
      seed = seed,
      A = A
    )
  }
  
  names(cluster_results) <- paste0("Cluster_", 1:best_c)
  
  list(
    best_c = best_c,
    cluster_results = cluster_results,
    cluster_sizes = cluster_sizes,
    cluster_assignments = as.integer(cluster_assignments),
    criteria = criteria_df
  )
}

#' Tune k in KNNRLD using Cross-Validation
#'
#' Finds the optimal number of neighbors for KNN regression in longitudinal data
#' using k-fold cross-validation. Evaluates k values from 2 to A.
#'
#' @param y Matrix of longitudinal outcomes (subjects x timepoints).
#' @param x Matrix of predictor variables (subjects x features).
#' @param nfolds Number of cross-validation folds (default = 10).
#' @param folds Optional list of pre-specified fold indices. If provided, nfolds is ignored.
#' @param seed Optional random seed for reproducibility.
#' @param A Maximum number of neighbors to evaluate (searches from 2 to A, default = 10).
#' @param graph Logical; if TRUE, plots MSPE vs. k.
#'
#' @return A list containing:
#' \item{crit}{Mean squared prediction error (MSPE) for each k value}
#' \item{best_k}{Optimal number of neighbors (minimizes MSPE)}
#' \item{performance}{Minimum MSPE value}
#' \item{runtime}{Elapsed computation time}
#'
#' @examples
#' \donttest{
#' set.seed(123)
#' n <- 30
#' T <- 3
#' d <- 2
#' x <- matrix(runif(n * d), nrow = n)
#' y <- matrix(rnorm(n * T), nrow = n)
#' tune_result <- KNNRLD.tune(
#'   y = y,
#'   x = x,
#'   nfolds = 3,
#'   A = 4
#' )
#' str(tune_result)
#' }
#'
#' @importFrom Directional makefolds
#' @importFrom Rfast colmeans
#' @importFrom graphics plot abline points
#'
#' @export
KNNRLD.tune <- function(y, x, nfolds = 10, folds = NULL, seed = NULL, A = 10, graph = FALSE) {
  
  y <- as.matrix(y)
  x <- as.matrix(x)
  n <- nrow(y)
  
  if (A < 2) {
    stop("A must be at least 2")
  }
  if (A > n) {
    warning(paste("A (", A, ") exceeds number of samples (", n, "). Reducing to", n))
    A <- n
  }
  
  if (is.null(folds)) {
    ina <- 1:n
    folds <- Directional::makefolds(ina, nfolds = nfolds, stratified = FALSE, seed = seed)
  }
  
  nfolds <- length(folds)
  k_values <- 2:A
  n_k <- length(k_values)
  
  fold_errors <- matrix(nrow = nfolds, ncol = n_k)
  
  runtime <- proc.time()
  
  for (fold_idx in 1:nfolds) {
    test_indices <- folds[[fold_idx]]
    train_indices <- setdiff(1:n, test_indices)
    
    y_test <- y[test_indices, , drop = FALSE]
    y_train <- y[train_indices, , drop = FALSE]
    x_test <- x[test_indices, , drop = FALSE]
    x_train <- x[train_indices, , drop = FALSE]
    
    predictions <- KNNRLD(xnew = x_test, y = y_train, x = x_train, k = k_values)
    
    for (j in 1:n_k) {
      fold_errors[fold_idx, j] <- mean((predictions[[j]] - y_test)^2, na.rm = TRUE)
    }
  }
  
  runtime <- proc.time() - runtime
  
  mspe <- Rfast::colmeans(fold_errors)
  names(mspe) <- paste0("k = ", k_values)
  
  best_k_idx <- which.min(mspe)
  best_k <- k_values[best_k_idx]
  best_performance <- mspe[best_k_idx]
  
  if (isTRUE(graph)) {
    graphics::plot(
      k_values, mspe,
      xlab = "Number of Nearest Neighbors (k)",
      ylab = "Mean Squared Prediction Error (MSPE)",
      type = "b",
      pch = 16,
      cex.axis = 1.2,
      cex.lab = 1.2,
      col = "darkgreen",
      lwd = 2,
      main = "KNN Tuning for Longitudinal Data"
    )
    graphics::abline(v = k_values, col = "lightgrey", lty = 2)
    graphics::abline(h = pretty(mspe), lty = 2, col = "lightgrey")
    graphics::points(best_k, best_performance, col = "red", pch = 19, cex = 1.5)
  }
  
  list(
    crit = mspe,
    best_k = best_k,
    performance = best_performance,
    runtime = runtime[3]
  )
}

#' Find Optimal Number of Clusters for Longitudinal Data
#'
#' This function determines the best number of clusters (C) for longitudinal data
#' clustering based on internal validation indices using the latrend package.
#'
#' @param Y A matrix or data frame of longitudinal outcomes (subjects x timepoints).
#' @param range_clusters A numeric vector of cluster numbers to evaluate (e.g., 2:6).
#' @param method Clustering method to use. Currently supports "kml" (default).
#'
#' @return A list with:
#' \item{best_c}{Optimal number of clusters}
#' \item{criteria}{Data frame of criteria values for each cluster number}
#' \item{criteria_best}{Criteria values for the best cluster number}
#'
#' @examples
#' \donttest{
#' set.seed(123)
#' n <- 30
#' T <- 3
#' y <- matrix(rnorm(n * T), nrow = n)
#' best_c_info <- BestC(Y = y, range_clusters = 2:4)
#' print(best_c_info$best_c)
#' }
#'
#' @importFrom latrend lcMethodKML latrend trajectoryAssignments
#' @importFrom stats reshape logLik AIC BIC
#'
#' @export
BestC <- function(Y, range_clusters = 2:6, method = "kml") {
  
  Y <- as.matrix(Y)
  n_subjects <- nrow(Y)
  n_timepoints <- ncol(Y)
  
  long_data <- data.frame(
    id = rep(1:n_subjects, each = n_timepoints),
    time = rep(1:n_timepoints, times = n_subjects),
    value = as.vector(t(Y))
  )
  
  crit_list <- list()
  
  calc_calinski <- function(data, clusters) {
    k <- length(unique(clusters))
    n <- nrow(data)
    if (k < 2 || n <= k) return(NA)
    overall_mean <- colMeans(data)
    between_ss <- 0
    for (i in 1:k) {
      cluster_data <- data[clusters == i, , drop = FALSE]
      if (nrow(cluster_data) > 0) {
        cluster_mean <- colMeans(cluster_data)
        between_ss <- between_ss + nrow(cluster_data) * sum((cluster_mean - overall_mean)^2)
      }
    }
    within_ss <- 0
    for (i in 1:k) {
      cluster_data <- data[clusters == i, , drop = FALSE]
      if (nrow(cluster_data) > 0) {
        cluster_mean <- colMeans(cluster_data)
        within_ss <- within_ss + sum(rowSums((cluster_data - cluster_mean)^2))
      }
    }
    if (within_ss == 0) return(NA)
    return((between_ss / (k - 1)) / (within_ss / (n - k)))
  }
  
  for (k in range_clusters) {
    if (method == "kml") {
      method_obj <- lcMethodKML(
        response = "value",
        time = "time",
        id = "id",
        nClusters = k,
        save.data = FALSE
      )
    } else {
      stop("Currently only 'kml' method is supported via latrend")
    }
    
    model <- tryCatch({
      latrend(method_obj, data = long_data, verbose = FALSE)
    }, error = function(e) {
      warning(paste("Clustering failed for k =", k, ":", e$message))
      return(NULL)
    })
    
    if (!is.null(model)) {
      cluster_assign <- trajectoryAssignments(model)
      traj_data <- as.data.frame(model)
      traj_wide <- stats::reshape(traj_data, 
                                  idvar = "id", 
                                  timevar = "time", 
                                  direction = "wide")
      traj_matrix <- as.matrix(traj_wide[, grep("value", names(traj_wide))])
      colnames(traj_matrix) <- NULL
      calinski_val <- calc_calinski(traj_matrix, cluster_assign)
      criteria <- c(
        CalinskiHarabasz = calinski_val,
        LogLik = as.numeric(stats::logLik(model)),
        AIC = stats::AIC(model),
        BIC = stats::BIC(model)
      )
      crit_list[[paste0("c", k)]] <- criteria
    } else {
      crit_list[[paste0("c", k)]] <- rep(NA, 4)
    }
  }
  
  max_len <- max(sapply(crit_list, length))
  crit_list_padded <- lapply(crit_list, function(x) {
    if (length(x) < max_len) c(x, rep(NA, max_len - length(x))) else x
  })
  
  criteria_df <- as.data.frame(do.call(cbind, crit_list_padded))
  rownames(criteria_df) <- c("CalinskiHarabasz", "LogLik", "AIC", "BIC")
  
  calinski_row <- which(rownames(criteria_df) == "CalinskiHarabasz")
  if (length(calinski_row) > 0 && any(!is.na(criteria_df[calinski_row, ]))) {
    best_c <- range_clusters[which.max(criteria_df[calinski_row, ])]
  } else {
    best_c <- range_clusters[1]
    warning("Could not determine optimal clusters. Using first value.")
  }
  
  criteria_best <- criteria_df[, paste0("c", best_c), drop = FALSE]
  
  list(
    best_c = as.integer(best_c),
    criteria = criteria_df,
    criteria_best = criteria_best
  )
}

#' Cluster-based KNN Regression for Longitudinal Data (CKNNRLD)
#'
#' This function implements a clustering-based KNN regression method designed
#' for longitudinal datasets. It first clusters the training data using longitudinal
#' clustering (via latrend) and then performs KNN regression within each selected cluster.
#'
#' @param xnew A matrix of predictor values for test data (new observations).
#' @param y A matrix or data frame of longitudinal responses (subjects x timepoints).
#' @param x A matrix or data frame of predictors for training data.
#' @param k Number of nearest neighbors to use. Can be a scalar (same k for all clusters)
#'          or a vector (different k per cluster).
#' @param c Number of clusters for longitudinal clustering.
#' @param cluster_method Clustering method to use. Currently supports "kml" (default).
#'
#' @return A data frame with predicted values and cluster assignment for each observation in `xnew`.
#'         Columns: cluster, Y1, Y2, ..., Yd (where d = number of timepoints).
#'
#' @examples
#' \donttest{
#' set.seed(123)
#' n <- 30
#' T <- 3
#' d <- 2
#' x <- matrix(runif(n * d), nrow = n)
#' y <- matrix(rnorm(n * T), nrow = n)
#' train_idx <- sample(1:n, 20)
#' test_idx <- setdiff(1:n, train_idx)
#' result <- CKNNRLD(
#'   x = x[train_idx, ],
#'   y = y[train_idx, ],
#'   xnew = x[test_idx, ],
#'   k = 3,
#'   c = 2
#' )
#' head(result)
#' }
#'
#' @importFrom latrend lcMethodKML latrend trajectoryAssignments
#' @importFrom Rfast dista colmeans
#'
#' @export
CKNNRLD <- function(xnew, y, x, k = 5, c = 4, cluster_method = "kml") {
  
  y <- as.matrix(y)
  x <- as.matrix(x)
  xnew <- as.matrix(xnew)
  
  if (nrow(y) != nrow(x)) {
    stop("Number of rows in y and x must be equal")
  }
  
  n_subjects <- nrow(y)
  n_timepoints <- ncol(y)
  n_predictors <- ncol(x)
  
  long_data <- data.frame(
    id = 1:n_subjects,
    time = rep(1:n_timepoints, each = n_subjects),
    value = as.vector(t(y))
  )
  
  if (cluster_method == "kml") {
    method_obj <- lcMethodKML(
      response = "value",
      time = "time",
      id = "id",
      nClusters = c,
      save.data = FALSE
    )
  } else {
    stop(paste("Method", cluster_method, "is not implemented. Use 'kml'."))
  }
  
  model <- latrend(method_obj, data = long_data, verbose = FALSE)
  cluster_assignments <- trajectoryAssignments(model)
  
  if (length(cluster_assignments) != n_subjects) {
    stop("Cluster assignment length doesn't match number of rows in y")
  }
  
  tdata <- cbind(cluster = cluster_assignments, y, x)
  
  y_start_idx <- 2
  y_end_idx <- y_start_idx + n_timepoints - 1
  x_start_idx <- y_end_idx + 1
  x_end_idx <- x_start_idx + n_predictors - 1
  
  cluster_centers <- matrix(NA, nrow = c, ncol = n_predictors)
  for (i in 1:c) {
    cluster_data <- tdata[tdata[, 1] == i, x_start_idx:x_end_idx, drop = FALSE]
    if (nrow(cluster_data) > 1) {
      cluster_centers[i, ] <- colMeans(cluster_data)
    } else {
      cluster_centers[i, ] <- cluster_data
    }
  }
  
  if (!identical(xnew, x)) {
    dist_result <- Rfast::dista(xnew, cluster_centers, trans = TRUE, k = 1, index = TRUE)
    xnew_clustered <- cbind(dist_result, xnew)
  } else {
    xnew_clustered <- cbind(cluster_assignments, xnew)
  }
  
  predictions <- vector("list", c)
  
  for (i in 1:c) {
    cluster_rows <- which(xnew_clustered[, 1] == i)
    
    if (length(cluster_rows) == 0) {
      predictions[[i]] <- matrix(NA, nrow = 0, ncol = n_timepoints)
      next
    }
    
    xnew_i <- xnew_clustered[cluster_rows, -1, drop = FALSE]
    if (nrow(xnew_i) == 1) {
      xnew_i <- matrix(xnew_i, nrow = 1)
    }
    
    y_i <- tdata[tdata[, 1] == i, y_start_idx:y_end_idx, drop = FALSE]
    x_i <- tdata[tdata[, 1] == i, x_start_idx:x_end_idx, drop = FALSE]
    
    k_used <- if (length(k) > 1) {
      min(k[i], nrow(y_i))
    } else {
      min(k, nrow(y_i))
    }
    
    knn_result <- KNNRLD(xnew = xnew_i, y = y_i, x = x_i, k = k_used)
    predictions[[i]] <- knn_result[[1]]
  }
  
  final_predictions <- matrix(NA, nrow = nrow(xnew), ncol = n_timepoints)
  for (i in 1:c) {
    cluster_rows <- which(xnew_clustered[, 1] == i)
    if (length(cluster_rows) > 0) {
      final_predictions[cluster_rows, ] <- predictions[[i]]
    }
  }
  
  result <- data.frame(cluster = xnew_clustered[, 1], final_predictions)
  colnames(result)[-1] <- paste0("Y", 1:n_timepoints)
  
  return(result)
}

#' Standard K-Nearest Neighbor Regression for Longitudinal Data
#'
#' This function performs KNN regression for longitudinal data without clustering.
#' It predicts longitudinal outcomes for new observations based on the weighted
#' average of their k nearest neighbors in the predictor space.
#'
#' @param xnew A matrix of predictor values for prediction (test set).
#' @param y A matrix or data frame of longitudinal responses (training set).
#' @param x A matrix or data frame of training predictor values.
#' @param k Number of nearest neighbors to use. Can be a scalar or a vector.
#'
#' @return A list of matrices with predicted values for each value of k.
#'         Each matrix has dimensions nrow(xnew) x ncol(y).
#'
#' @examples
#' \donttest{
#' set.seed(123)
#' n <- 30
#' T <- 3
#' d <- 2
#' x <- matrix(runif(n * d), nrow = n)
#' y <- matrix(rnorm(n * T), nrow = n)
#' train_idx <- sample(1:n, 20)
#' test_idx <- setdiff(1:n, train_idx)
#' pred <- KNNRLD(
#'   xnew = x[test_idx, ],
#'   y = y[train_idx, ],
#'   x = x[train_idx, ],
#'   k = 3
#' )
#' head(pred[[1]])
#' }
#'
#' @importFrom Rfast dista colmeans
#'
#' @export
KNNRLD <- function(xnew, y, x, k = 5) {
  
  y <- as.matrix(y)
  x <- as.matrix(x)
  xnew <- as.matrix(xnew)
  
  n_timepoints <- ncol(y)
  n_test <- nrow(xnew)
  k_values <- sort(k)
  n_k <- length(k_values)
  max_k <- max(k_values)
  
  if (max_k > nrow(x)) {
    warning(paste("Maximum k (", max_k, ") exceeds number of training samples (", 
                  nrow(x), "). Reducing to ", nrow(x)))
    max_k <- nrow(x)
    k_values <- k_values[k_values <= max_k]
    n_k <- length(k_values)
  }
  
  distance_result <- Rfast::dista(xnew, x, trans = FALSE, k = max_k, index = TRUE)
  
  predictions <- list()
  
  for (j in 1:n_k) {
    current_k <- k_values[j]
    pred_matrix <- matrix(NA, nrow = n_test, ncol = n_timepoints)
    
    for (i in 1:n_test) {
      neighbor_indices <- distance_result[1:current_k, i]
      pred_matrix[i, ] <- Rfast::colmeans(y[neighbor_indices, , drop = FALSE])
    }
    
    predictions[[j]] <- pred_matrix
  }
  
  names(predictions) <- paste("k =", k_values)
  
  if (n_k == 1) {
    return(predictions)
  }
  
  return(predictions)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BestC.R
\name{BestC}
\alias{BestC}
\title{Find Optimal Number of Clusters for Longitudinal Data}
\usage{
BestC(Y, range_clusters = 2:6, method = "kml")
}
\arguments{
\item{Y}{A matrix or data frame of longitudinal outcomes (subjects x timepoints).}

\item{range_clusters}{A numeric vector of cluster numbers to evaluate (e.g., 2:6).}

\item{method}{Clustering method to use. Currently supports "kml" (default).}
}
\value{
A list with:
\item{best_c}{Optimal number of clusters}
\item{criteria}{Data frame of criteria values for each cluster number}
\item{criteria_best}{Criteria values for the best cluster number}
}
\description{
This function determines the best number of clusters (C) for longitudinal data
clustering based on internal validation indices using the latrend package.
}
\examples{
\donttest{
set.seed(123)
n <- 30
T <- 3
y <- matrix(rnorm(n * T), nrow = n)
best_c_info <- BestC(Y = y, range_clusters = 2:4)
print(best_c_info$best_c)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tuning_CKNNRLD.R
\name{CKNNRLD.tune}
\alias{CKNNRLD.tune}
\title{Tune CKNNRLD Model with Automatic Cluster Selection}
\usage{
CKNNRLD.tune(
  y,
  x,
  nfolds = 10,
  folds = NULL,
  seed = NULL,
  A = 10,
  C_range = 2:6,
  cluster_method = "kml"
)
}
\arguments{
\item{y}{Matrix of longitudinal outcomes (subjects x timepoints).}

\item{x}{Matrix of predictor variables (subjects x features).}

\item{nfolds}{Number of folds for cross-validation (default = 10).}

\item{folds}{Optional list of pre-specified fold indices.}

\item{seed}{Random seed for reproducibility.}

\item{A}{Maximum number of neighbors to evaluate (searches from 2 to A, default = 10).}

\item{C_range}{Range of cluster numbers to evaluate (default = 2:6).}

\item{cluster_method}{Clustering method to use. Currently supports "kml" (default).}
}
\value{
A list containing:
\item{best_c}{Optimal number of clusters}
\item{cluster_results}{Tuning results for each cluster (from KNNRLD.tune)}
\item{cluster_sizes}{Size of each cluster}
\item{cluster_assignments}{Cluster labels for each subject}
\item{criteria}{Data frame of quality criteria used for selecting best C}
}
\description{
Automatically selects the best number of clusters (C) using internal validation
criteria, then tunes the CKNNRLD model within each cluster via cross-validation.
}
\examples{
\donttest{
set.seed(123)
n <- 30
T <- 3
d <- 2
x <- matrix(runif(n * d), nrow = n)
y <- matrix(rnorm(n * T), nrow = n)
tune_result <- CKNNRLD.tune(
  y = y,
  x = x,
  nfolds = 3,
  A = 4,
  C_range = 2:3
)
print(tune_result$best_c)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KNNRLD.R
\name{KNNRLD}
\alias{KNNRLD}
\title{Standard K-Nearest Neighbor Regression for Longitudinal Data}
\usage{
KNNRLD(xnew, y, x, k = 5)
}
\arguments{
\item{xnew}{A matrix of predictor values for prediction (test set).}

\item{y}{A matrix or data frame of longitudinal responses (training set).}

\item{x}{A matrix or data frame of training predictor values.}

\item{k}{Number of nearest neighbors to use. Can be a scalar or a vector.}
}
\value{
A list of matrices with predicted values for each value of k.
Each matrix has dimensions nrow(xnew) x ncol(y).
}
\description{
This function performs KNN regression for longitudinal data without clustering.
It predicts longitudinal outcomes for new observations based on the weighted
average of their k nearest neighbors in the predictor space.
}
\examples{
\donttest{
set.seed(123)
n <- 30
T <- 3
d <- 2
x <- matrix(runif(n * d), nrow = n)
y <- matrix(rnorm(n * T), nrow = n)
train_idx <- sample(1:n, 20)
test_idx <- setdiff(1:n, train_idx)
pred <- KNNRLD(
  xnew = x[test_idx, ],
  y = y[train_idx, ],
  x = x[train_idx, ],
  k = 3
)
head(pred[[1]])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CKNNRLD.R
\name{CKNNRLD}
\alias{CKNNRLD}
\title{Cluster-based KNN Regression for Longitudinal Data (CKNNRLD)}
\usage{
CKNNRLD(xnew, y, x, k = 5, c = 4, cluster_method = "kml")
}
\arguments{
\item{xnew}{A matrix of predictor values for test data (new observations).}

\item{y}{A matrix or data frame of longitudinal responses (subjects x timepoints).}

\item{x}{A matrix or data frame of predictors for training data.}

\item{k}{Number of nearest neighbors to use. Can be a scalar (same k for all clusters)
or a vector (different k per cluster).}

\item{c}{Number of clusters for longitudinal clustering.}

\item{cluster_method}{Clustering method to use. Currently supports "kml" (default).}
}
\value{
A data frame with predicted values and cluster assignment for each observation in \code{xnew}.
Columns: cluster, Y1, Y2, ..., Yd (where d = number of timepoints).
}
\description{
This function implements a clustering-based KNN regression method designed
for longitudinal datasets. It first clusters the training data using longitudinal
clustering (via latrend) and then performs KNN regression within each selected cluster.
}
\examples{
\donttest{
set.seed(123)
n <- 30
T <- 3
d <- 2
x <- matrix(runif(n * d), nrow = n)
y <- matrix(rnorm(n * T), nrow = n)
train_idx <- sample(1:n, 20)
test_idx <- setdiff(1:n, train_idx)
result <- CKNNRLD(
  x = x[train_idx, ],
  y = y[train_idx, ],
  xnew = x[test_idx, ],
  k = 3,
  c = 2
)
head(result)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Tuning_KNNRLD.R
\name{KNNRLD.tune}
\alias{KNNRLD.tune}
\title{Tune k in KNNRLD using Cross-Validation}
\usage{
KNNRLD.tune(
  y,
  x,
  nfolds = 10,
  folds = NULL,
  seed = NULL,
  A = 10,
  graph = FALSE
)
}
\arguments{
\item{y}{Matrix of longitudinal outcomes (subjects x timepoints).}

\item{x}{Matrix of predictor variables (subjects x features).}

\item{nfolds}{Number of cross-validation folds (default = 10).}

\item{folds}{Optional list of pre-specified fold indices. If provided, nfolds is ignored.}

\item{seed}{Optional random seed for reproducibility.}

\item{A}{Maximum number of neighbors to evaluate (searches from 2 to A, default = 10).}

\item{graph}{Logical; if TRUE, plots MSPE vs. k.}
}
\value{
A list containing:
\item{crit}{Mean squared prediction error (MSPE) for each k value}
\item{best_k}{Optimal number of neighbors (minimizes MSPE)}
\item{performance}{Minimum MSPE value}
\item{runtime}{Elapsed computation time}
}
\description{
Finds the optimal number of neighbors for KNN regression in longitudinal data
using k-fold cross-validation. Evaluates k values from 2 to A.
}
\examples{
\donttest{
set.seed(123)
n <- 30
T <- 3
d <- 2
x <- matrix(runif(n * d), nrow = n)
y <- matrix(rnorm(n * T), nrow = n)
tune_result <- KNNRLD.tune(
  y = y,
  x = x,
  nfolds = 3,
  A = 4
)
str(tune_result)
}

}
